package TM::Serializable::XTM;
# $Id:$ 
use Class::Trait 'base';
use Class::Trait 'TM::Serializable';
use Data::Dumper;
use XML::Writer;
use XML::LibXML;
use IO::String;
use TM::Literal;
use strict;

use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);
$VERSION = qw(('$Revision:$'))[1];

=pod

=head1 NAME

TM::Serializable::XTM - Topic Maps, trait for parsing and dumping XTM instances.

=head1 SYNOPSIS

  # this is not an end-user package
  # see the source in TM::Materialized::XTM how this can be used

=head1 DESCRIPTION

This trait provides parsing and dumping functionality for XTM 1.0 instances. 

=head1 INTERFACE

=head2 Methods

=over

=item B<deserialize>

This method takes an XTM string and tries to parse it. It will raise 
an exception on parse error.

=cut

sub deserialize {
    my ($self,$content)=@_;

    my $xp=XML::LibXML->new();
    # xpath and namespace horror, explained at
    # http://search.cpan.org/~pajas/XML-LibXML-1.65/lib/XML/LibXML/Node.pod
    # at findnodes. in short: default ns declaration fucks up everything
    # and xpath findnodes() no longer works, as you have no prefix to
    # feed it, and you must scream, because the nodes belong to  
    # a ns - just one whose name you can't tell fucking xpath...
    $content=~s!<topicMap +xmlns="http://www.topicmaps.org/xtm/1.0/"!<topicMap !;
    my $doc=$xp->parse_string($content);

    my $base=$self->baseuri;
    my %reifiedby; # what by whom, for assocs
    for my $t ($doc->findnodes('/topicMap/topic'))
    {
	my ($id,$sloc,@sin);
	$id=$t->findvalue('@id');
	for ($t->findnodes("subjectIdentity/subjectIndicatorRef"))
	{
	    push @sin,$_->findvalue('./@xlink:href');
	}
	$sloc=$t->findvalue('subjectIdentity/resourceRef/@xlink:href');
	my $locretar=$t->findvalue('subjectIdentity/topicRef/@xlink:href');
	die "topic $id: standards violation, duplicate subject locators ($sloc and $locretar)\n"
	    if ($sloc && $locretar);
	if ($locretar)		# this topic reifies another topic or assoc in this map
	{
	    $locretar=~s/^\#//;
	    if ($doc->findnodes("/topicMap/topic[\@id='$locretar']"))
	    {
		$sloc=$base.$locretar;
	    }
	    elsif ($doc->findnodes("/topicMap/association[\@id='$locretar']"))
	    {
		# we don't know the final association id until we handle them,
		# so resolution of this reification must wait until then.
		$reifiedby{$locretar}=$id; 
	    }
	    else
	    {
		die "topic $id: reifies nonexistent other topic ($locretar)\n";
	    }
	}
	$self->internalize($base.$id=>$sloc);
	map { $self->internalize($base.$id=> \$_); } (@sin);

	for my $bn ($t->findnodes("baseName"))
	{
	    my $scope=$bn->findvalue('scope/topicRef/@xlink:href');
	    $scope=~s/^\#//;
	    $self->assert(Assertion->new(kind=>TM->NAME,
					 type=>'name',
					 roles=>[qw(thing value)],
					 players=>[$id, 
						   TM::Literal->new($bn->findvalue("baseNameString/text()"),
								    TM::Literal->STRING)],
					 scope=>$scope));
	}

	for my $oc ($t->findnodes("occurrence"))
	{
	    my $scope=$oc->findvalue('scope/topicRef/@xlink:href');
	    $scope=~s/^\#//;
	    my $type=$oc->findvalue('instanceOf/topicRef/@xlink:href');
	    $type=~s/^\#//;
	    $type||="occurrence";
	    my $isin=$oc->findvalue('resourceData/text()');
	    my $value=$isin||$oc->findvalue('resourceRef/@xlink:href');

	    $self->assert(Assertion->new(kind=>TM->OCC,
					 type=>$type,
					 roles=>[qw(thing value)],
					 players=>[$id, 
						   TM::Literal->new($value,
								    $isin? TM::Literal->STRING : TM::Literal->URI)],
					 scope=>$scope));
	}
	for my $p ($t->findnodes('instanceOf'))
	{
	    my $pt=$p->findvalue('topicRef/@xlink:href');
	    $pt=~s/^\#//;
	    $self->assert(Assertion->new(kind=>TM->ASSOC,
					 type=>"isa",
					 roles=>[qw(class instance)],
					 players=>[$pt,$id]));
	}
    }   

    for my $a ($doc->findnodes('/topicMap/association'))
    {
	my $id=$a->findvalue('@id');
	my $type=$a->findvalue('instanceOf/topicRef/@xlink:href');
	$type=~s/^\#//;
	my $scope=$a->findvalue('scope/topicRef/@xlink:href');
	$scope=~s/^\#//;
	my (@roles,@players);

	for my $m ($a->findnodes('member'))
	{
	    my $role=$m->findvalue('roleSpec/topicRef/@xlink:href');
	    $role=~s/^\#//;
	    for my $p ($m->findnodes('topicRef'))
	    {
		my $player=$p->findvalue('@xlink:href');
		$player=~s/^\#//;
		push @roles,$role;
		push @players,$player;
	    }
	}
	my $newass=($self->assert(Assertion->new(kind=>TM->ASSOC,
						 type=>$type,
						 scope=>$scope,
						 roles=>\@roles,
						 players=>\@players)))[0];
	my $lid=$newass->[TM->LID];
	
	# now resolve reification of this particular assoc 
	if ($reifiedby{$id})
	{
	    $self->internalize($base.$reifiedby{$id}=>$lid);
	    delete $reifiedby{$id};
	}
    }
    die "map contains invalid reification info for topics ".join(", ",values %reifiedby)."\n"
	if (%reifiedby);
}

=pod

=item B<serialize>

This method serializes the map object into XTM notation and returns the resulting string.  
It will raise an exception if the object contains constructs that XTM cannot represent. 
The result is a standard Perl string, so you may need to force it into a particular encoding.

The method understands a number of key/value pair parameters:

=over

=item C<omit_trivia> (default: C<0>)

This option suppresses the output of completely "naked" topics (topics without any characteristics).

=item C<version> (default: C<1>)

This option controls whether XTM 1.0 or XTM 2.0 is generated. Currently only XTM 1.0 is supported.

=back

=cut

sub serialize  
{
    my ($self,%opts)=@_;
    $opts{version}||=1;

    die(scalar __PACKAGE__ .": serialization unimplemented for XTM version "
	.$opts{version}) if ($opts{version} ne 1);
        
    my $base=$self->{baseuri};
    # the work starts: collect info from the assertions.
    my (%topics,%assocs);
    for my $m ($self->match(TM->FORALL))
    {
	my $kind=$m->[TM->KIND];
	my $type=$m->[TM->TYPE];
	my $scope=$m->[TM->SCOPE];
	$scope=~s/^$base//;
	undef $scope if ($scope eq "us");
	$type=~s/^$base//;
	my $lid=$m->[TM->LID];

	if ($kind==TM->ASSOC) 
	{
	    if ($type eq "isa")
	    {
		my ($p,$c);
		$p=($self->get_x_players($m,$base."class"))[0];
		$c=($self->get_x_players($m,$base."instance"))[0];
		$p=~s/^$base//;
		$c=~s/^$base//;
		push @{$topics{$p}->{children}},$c;
		push @{$topics{$c}->{parents}},$p;
	    }
	    else
	    {
		my %thisa;
		$thisa{type}=$type;
		$thisa{scope}=$scope;
		die "XTM 1.x does not offer reification by associations\n"
		    if ($thisa{reifies});
		
		for my $role (@{$self->get_role_s($m)})
		{
		    my $rolename=$role;
		    $rolename=~s/^$base//;
		    # must prime the array...
		    $thisa{roles}->{$rolename}=[];
		    
		    for my $player ($self->get_x_players($m,$role))
		    {
			$player=~s/^$base//;
			push @{$thisa{roles}->{$rolename}}, $player;
		    }
		}
		$assocs{$lid}=\%thisa;
	    }
	}
	elsif ($kind==TM->NAME)
	{
	    my $name=($self->get_x_players($m,$base."thing"))[0];
	    $name=~s/^$base//;
	    undef $type if ($type eq "name");
	    
	    die "XTM 1.0 does not offer reification of basenames (topic $name)\n"
		if ($self->midlet($lid)->[TM->ADDRESS]);
	    
	    for my $p ($self->get_x_players($m,$base."value"))
	    {
		next if (!($p->[0]));
		push @{$topics{$name}->{bn}}, [ $p->[0], $type, $scope ];
	    }
	}
	elsif ($kind==TM->OCC)
	{
	    my $key="oc";
	    my $name=($self->get_x_players($m,$base."thing"))[0];
	    $name=~s/^$base//;
	    
	    die "XTM 1.x does not offer reification of occurrences (topic $name)\n"
		if ($self->midlet($lid)->[TM->ADDRESS]);
	    
	    undef $type if ($type eq "occurrence");
	    for my $p ($self->get_x_players($m,$base."value"))
	    { 
		next if (!($p->[0]));
		$key="in" if ($p->[1] ne TM::Literal->URI);
		push @{$topics{$name}->{$key}}, [ $p->[0], $type, $scope ];
	    }
	}
    }
    
    # then from the topics
    # we also need to run this part because of all the reification-crap...
    # uuuggly distinction between topics and assertions
    for my $t (grep(!$self->retrieve($_), $self->midlets))
    {
	my $tn=$t;
	$tn=~s/^$base//; 
	$topics{$tn}||={}; 
	
	die "variants are not implemented (topic $tn)\n"
	    if ($self->variants($t));
	
	my $reifies=$self->midlet($t)->[TM->ADDRESS];
	if ($reifies)
	{
	    # xtm 1.0: reification is always attached to the active topic
	    if ($self->retrieve($reifies))	
	    {
		# assocs: we need to give them xml-compliant ids and point to them
		$reifies=~s/^$base/\#_/;
	    }
	    elsif ($reifies=~/^$base(.+)$/)
	    {
		$reifies="#".$1;
	    }
	    $topics{$tn}->{reifies}=$reifies; 
	}
	my $sin=$self->midlet($t)->[TM->INDICATORS];
	if ($sin && @$sin)
	{
	    $topics{$tn}->{sin}=$sin;
	}
    } 
    
    # finally the actual dumping of the information
    my $ios = IO::String->new;
    my $xlns='http://www.w3.org/1999/xlink';
    my $xtmns="http://www.topicmaps.org/xtm/1.0/";
    my $writer = new XML::Writer(OUTPUT=>$ios,NAMESPACES=>1,
				 PREFIX_MAP=>{$xtmns=>"",
					      $xlns=>"xlink"},
				 FORCED_NS_DECLS=>[$xtmns,$xlns],
				 NEWLINES=>1);
    $writer->xmlDecl("iso-8859-1");
    $writer->startTag("topicMap");

    for my $t (sort keys %topics)
    {
	my $tn=$topics{$t};
	next if ($opts{omit_trivia} && !keys %$tn);

	$writer->startTag("topic","id"=>$t);
	for my $p (@{$tn->{parents}})
	{
	    $writer->startTag("instanceOf");
	    $writer->emptyTag("topicRef",[$xlns,"href"]=>"#$p");
	    $writer->endTag;
	}
	if ($tn->{reifies} || $tn->{sin})
	{
	    $writer->startTag("subjectIdentity");

	    if ($tn->{reifies})
	    {
		# reifies a topic/assoc within this map? then do an internal topicref
		$writer->emptyTag(($tn->{reifies}=~/\#/?"topicRef":"resourceRef"),
				      [$xlns,"href"]=>$tn->{reifies}); 
	    }
	    for my $n (@{$tn->{sin}})
	    {
		$writer->emptyTag("subjectIndicatorRef",[$xlns,"href"]=>$n);
	    }
	    $writer->endTag;
	}

	if ($tn->{bn})
	{
	    for my $n (@{$tn->{bn}})
	    {
		$writer->startTag("baseName");
		if ($n->[2])
		{
		    $writer->startTag("scope");
		    $writer->emptyTag("topicRef",[$xlns,"href"]=>"#".$n->[2]);
		    $writer->endTag;
		}
		$writer->dataElement("baseNameString",$n->[0]);
		$writer->endTag;
	    }
	}
	for my $k (qw(in oc))
	{
	    if ($tn->{$k})
	    {
		for my $n (@{$tn->{$k}})
		{
		    $writer->startTag("occurrence");
		    if ($n->[1])
		    {
			$writer->startTag("instanceOf");
			$writer->emptyTag("topicRef",[$xlns,"href"]=>"#".$n->[1]);
			$writer->endTag;
		    }
		    if ($n->[2])
		    {
			$writer->startTag("scope");
			$writer->emptyTag("topicRef",[$xlns,"href"]=>"#".$n->[2]);
			$writer->endTag;
		    }
		    if ($k eq "in")
		    {
			$writer->dataElement("resourceData",$n->[0]);
		    }
		    else
		    {
			$writer->emptyTag("resourceRef",[$xlns,"href"]=>$n->[0]);
		    }
		    $writer->endTag;
		}
	    }
	}
	$writer->endTag;
    }

    for (sort {$assocs{$b}->{type} cmp $assocs{$a}->{type}} keys %assocs)
    {
	my $a=$assocs{$_};
	my $n=$_;
	$n=~s/^$base/_/;
	$writer->startTag("association",id=>$n); # the id attribute is required for reification in xtm 1.0
	$writer->startTag("instanceOf");
	$writer->emptyTag("topicRef",[$xlns,"href"]=>"#".$a->{type});
	$writer->endTag;

	if ($a->{scope})
	{
	    $writer->startTag("scope");
	    $writer->emptyTag("topicRef",[$xlns,"href"]=>"#".$a->{scope});
	    $writer->endTag;
	}
	
	for my $r (sort keys %{$a->{roles}})
	{ 
	    $writer->startTag("member");
	    $writer->startTag("roleSpec");
	    $writer->emptyTag("topicRef",[$xlns,"href"]=>"#".$r);
	    $writer->endTag;	    
	    
	    map { $writer->emptyTag("topicRef",[$xlns,"href"]=>"#$_") } (@{$a->{roles}->{$r}});
	    $writer->endTag;
	}
	$writer->endTag;

    }
    
    $writer->endTag;
    $writer->end();
    return ${$ios->string_ref}; 
}

=pod

=back

=head1 SEE ALSO

L<TM>, L<TM::Serializable>

=head1 AUTHOR INFORMATION

Copyright 2007 Alexander Zangerl.

This library is free software; you can redistribute it and/or modify it under the same terms as Perl
itself.  http://www.perl.com/perl/misc/Artistic.html

=cut

1;

